#include "p9p.h"

enum {
	Suser,
	Snice,
	Ssystem,
	Sidle,
	Siowait,
	Sirq,
	Ssoftirq,
	Nstats,
};

enum{
	/* 4GB io default */
	Dbs = 1<<20,
	Dcount = 1<<12,

	USER_HZ = 100,
};

char *snames[] = {
[Suser] 	= "user",
[Snice] 	= "nice",
[Ssystem] 	= "system",
[Sidle] 	= "idle",
[Siowait]	= "iowait",
[Sirq]		= "irq",
[Ssoftirq]	= "softirq"
};

ulong rdstats[Nstats];
ulong wrstats[Nstats];
char file[4096];
char *block;
int fd;
vlong bs = Dbs;
vlong count = Dcount;
int fflag, ncpu;

void
sysfatal(char *fmt, ...)
{
	va_list arg;

	va_start(arg, fmt);
	fprintf(stderr, fmt, arg);
	va_end(arg);
	exit(1);
}

static char*
qtoken(char *s, char *sep)
{
	int quoting;
	char *t;

	quoting = 0;
	t = s;	/* s is output string, t is input string */
	while (*t != 0 && (quoting || strchr(sep, *t) == nil)) {
		if (*t != '\'') {
			*s++ = *t++;
			continue;
		}
		/* *t is a quote */
		if (!quoting) {
			quoting = 1;
			t++;
			continue;
		}
		/* quoting and we're on a quote */
		if (t[1] != '\'') {
			/* end of quoted section; absorb closing quote */
			t++;
			quoting = 0;
			continue;
		}
		/* doubled quote; fold one quote into two */
		t++;
		*s++ = *t++;
	}
	if (*s != '\0') {
		*s = '\0';
		if(t == s)
			t++;
	}
	return t;
}

int
tokenize(char *s, char **args, int maxargs)
{
	int nargs;
	static char qsep[] = " \t\r\n";

	for (nargs=0; nargs<maxargs; nargs++) {
		while (*s != '\0' && strchr(qsep, *s))
			s++;
		if (*s == '\0')
			break;
		args[nargs] = s;
		s = qtoken(s, qsep);
	}

	return nargs;
}

void
procstat(ulong stats[Nstats])
{
	int i, n, fd, argc;
	char buf[128], *p, *argv[16], **args;

	args = argv;
	fd = open("/proc/stat", O_RDONLY);
	if (fd < 0)
		sysfatal("procstat open failure: %r.");
	n = read(fd, buf, sizeof buf - 1);
	if (n < 0)
		sysfatal("procstat read failure: %r.");
	buf[n] = '\0';
	p = strchr(buf, '\n');
	if (p == nil)
		sysfatal("procstat can't find eol in '%s'.", buf);
	*p = '\0';
	argc = tokenize(buf, argv, nelem(argv));
	if (argc < Nstats + 1)
		sysfatal("procstat format assumption failure.");
	args++;
	for (i=0; i<Nstats; i++)
		stats[i] = atoi(args[i]) - stats[i];
	close(fd);
}

void
ddr(void)
{
	int i, n;

	if (lseek(fd, 0, SEEK_SET) == (off_t)-1)
		sysfatal("unable to seek to zero for read.");

	fprintf(stderr, "Reading from %s ... ", file);
	procstat(rdstats);
	for (i=0; i<count; i++) {
		n = read(fd, block, bs);
		if (n != bs)
			sysfatal("short read %d, block %d", n, i);
	}
	procstat(rdstats);
	fprintf(stderr, "done.\n");
}

void
ddw(void)
{
	int i, n;

	if (lseek(fd, 0, SEEK_SET) == -1)
		sysfatal("unable to seek to zero for write.");

	fprintf(stderr, "Writing to %s ... ", file);
	procstat(wrstats);
	for (i=0; i<count; i++) {
		n = write(fd, block, bs);
		if (n != bs)
			sysfatal("short write %d, block %d", n, i);
	}
	fprintf(stderr, "syncing ... ");
	fsync(fd);
	procstat(wrstats);
	fprintf(stderr, "done.\n");
}

void
raw(ulong stats[Nstats])
{
	int i;

	for (i=0; i<Nstats; i++)
		printf("%s=%ld ", snames[i], stats[i]);
	printf("\n");
}

void
cooked(ulong stats[Nstats])
{
	vlong kbs, m, n;
	ulong cpuu;
	int i;

	for (i=0, n=100; i<Nstats; i++)
		n += stats[i];
	kbs = (bs * count);
	kbs /= n;
	kbs *= USER_HZ * ncpu;
	kbs >>= 10;

	m = stats[Sidle] + stats[Siowait];
	cpuu = (n - m) * 100 / n;

	printf("%6lld %6ld\n", kbs, cpuu);
}

void
ddt(void)
{
	printf("%-5lldMB  KiB/s    CPU%% \n", (bs*count) >> 20LL);
	printf("Write   ");
	cooked(wrstats);
	printf("Read    ");
	cooked(rdstats);
	printf("\n");
	printf("WRaw ");
	raw(wrstats);
	printf("RRaw ");
	raw(rdstats);
}

int
countcpus(void)
{
	char buf[4096], *p, *r;
	int fd, n;

	fd = open("/proc/stat", O_RDONLY);
	if (fd < 0)
		sysfatal("cpus: can't open /proc/stat");
	n = read(fd, buf, sizeof buf - 1);
	if (n < 0)
		sysfatal("cpus: can't read /proc/stat");
	buf[n] = 0;

	n = -1;
	for (p=r=buf; (r=strchr(p, '\n')); p=r) {
		*r++ = '\0';
		if (strncmp(p, "cpu", 3) == 0)
			n++;
	}
	return n;
}

void
usage(void)
{
	fprintf(stderr, "usage: ddt [-?] [-c count] [-b bs] dir\n"); 
	exit(1);
}


int
main(int argc, char *argv[])
{
	char *dir;
	int n;

	ARGBEGIN {
	case 'b':
		bs = atoi(EARGF(usage()));
		break;
	case 'c':
		count = atoi(EARGF(usage()));
		break;
	case 'f':
		strcpy(file, EARGF(usage()));
		fflag++;
		break;
	case '?':
		usage();
	} ARGEND

	if (argc != 1)
		usage();

	if (fflag == 0) {
		dir = *argv;
		n = strlen(dir);
		if (dir[--n] == '/')
			dir[n] = '\0';
		snprintf(file, sizeof file, "%s/ddt.%d", dir, getpid());
		fd = open(file, O_CREAT|O_RDWR|O_LARGEFILE|O_NOATIME, 0777);
		unlink(file);
	} else
		fd = open(file, O_RDWR|O_LARGEFILE|O_NOATIME);

	if (fd == -1)
		sysfatal("can't open %s: %r.", file);

	ncpu = countcpus();
	block = malloc(bs);
	if (block == nil)
		sysfatal("can't malloc %d size block for I/O.", bs);
	ddw();
	fprintf(stderr, "sleeping 10 seconds ... ");
	sleep(10);
	printf("done.\n");
	ddr();
	ddt();
	return 0;
}

